import random
import numpy as np
import matplotlib.pyplot as plt

#set line width
plt.rcParams['lines.linewidth'] = 4
#set font size for titles
plt.rcParams['axes.titlesize'] = 20
#set font size for labels on axes
plt.rcParams['axes.labelsize'] = 20
#set size of numbers on x-axis
plt.rcParams['xtick.labelsize'] = 16
#set size of numbers on y-axis
plt.rcParams['ytick.labelsize'] = 16
#set size of ticks on x-axis
plt.rcParams['xtick.major.size'] = 7
#set size of ticks on y-axis
plt.rcParams['ytick.major.size'] = 7
#set size of markers, e.g., circles representing points
#set numpoints for legend
plt.rcParams['legend.numpoints'] = 1
   
class Location_1D(object):
    def __init__(self, x):
        """x and y are numbers"""
        self.x = x

    def move(self, deltaX):
        """deltaX and deltaY are numbers"""
        return Location_1D(self.x + deltaX)
    
    def getX(self):
        return self.x

    def distFrom(self, other):
        return abs(self.x - other.getX())
   
class Location_2D(object):
    def __init__(self, x, y):
        """x and y are numbers"""
        self.x = x
        self.y = y

    def move(self, deltaX, deltaY):
        """deltaX and deltaY are numbers"""
        return Location_2D(self.x + deltaX, self.y + deltaY)
    
    def getX(self):
        return self.x

    def getY(self):
        return self.y

    def distFrom(self, other):
        return ((self.x - other.getX())**2 +
                (self.y - other.getY())**2)**0.5
   
class Location_3D(object):
    def __init__(self, x, y, z):
        """x and y are numbers"""
        self.x = x
        self.y = y
        self.z = z

    def move(self, deltaX, deltaY, deltaZ):
        """deltaX and deltaY are numbers"""
        return Location_3D(self.x + deltaX, self.y + deltaY,
                        self.z + deltaZ)
    def getX(self):
        return self.x

    def getY(self):
        return self.y
    
    def getZ(self):
        return self.z

    def distFrom(self, other):
        xDist = self.x - other.getX()
        yDist = self.y - other.getY()
        zDist = self.z - other.getZ()
        return (xDist**2 + yDist**2 + zDist**2)**0.5
    
class Field(object):
    def __init__(self):
        self.drunks = {}

    def addDrunk(self, drunk, loc):
        if drunk in self.drunks:
            raise ValueError('Duplicate drunk')
        else:
            self.drunks[drunk] = loc

    def getLoc(self, drunk):
        if drunk not in self.drunks:
            raise ValueError('Drunk not in field')
        return self.drunks[drunk]

class Field_1D(Field):

    def moveDrunk(self, drunk):
        if drunk not in self.drunks:
            raise ValueError('Drunk not in field')
        xDist = drunk.takeStep()
        #use move method of Location to get new location
        self.drunks[drunk] =\
            self.drunks[drunk].move(xDist)

class Field_2D(Field):

    def moveDrunk(self, drunk):
        if drunk not in self.drunks:
            raise ValueError('Drunk not in field')
        xDist, yDist = drunk.takeStep()
        #use move method of Location to get new location
        self.drunks[drunk] =\
            self.drunks[drunk].move(xDist, yDist)

class Field_3D(Field):

    def moveDrunk(self, drunk):
        if drunk not in self.drunks:
            raise ValueError('Drunk not in field')
        xDist, yDist, zDist = drunk.takeStep()
        #use move method of Location to get new location
        self.drunks[drunk] =\
            self.drunks[drunk].move(xDist, yDist, zDist)

class Drunk(object):
    def __init__(self, name = None):
        """Assumes name is a str"""
        self.name = name

    def __str__(self):
        if self != None:
            return self.name
        return 'Anonymous'
    
class Drunk_1D(Drunk):
    def takeStep(self):
        if random.random() < 0.5:
            return -1
        else:
            return 1

class Drunk_2D(Drunk):
    def takeStep(self):
        deltaX = random.random()
        deltaY = 1 - deltaX
        if random.random() < 0.5:
            deltaX = -deltaX
        if random.random() < 0.5:
            deltaY = -deltaY
        return deltaX, deltaY

class Drunk_3D(Drunk):
    def takeStep(self):
        deltaX = random.random()
        deltaY = random.uniform(deltaX, 1)
        deltaY -= deltaX
        deltaZ = 1 - deltaX - deltaY
        if random.random() < 0.5:
            deltaX = -deltaX
        if random.random() < 0.5:
            deltaY = -deltaY
        if random.random() < 0.5:
            deltaZ = -deltaZ
        return deltaX, deltaY, deltaZ

def walk(f, d, numSteps):
    """Assumes: f a Field, d a Drunk in f,
                and numSteps an int >= 0.
       Moves d numSteps times, and returns the
       distance between the final location and
       the location at the start of the walk."""
    start = f.getLoc(d)
    for s in range(numSteps):
        f.moveDrunk(d)
    return start.distFrom(f.getLoc(d))

def simWalks(numSteps, numTrials, dClass):
    """Assumes numSteps an int >= 0, numTrials an int > 0,
         dClass a subclass of Drunk
       Simulates numTrials walks of numSteps steps each.
       Returns a list of the final distances for each trial"""
    molecule = dClass('molecule')
    if dClass is Drunk_3D:
        origin = Location_3D(0, 0, 0)
    elif dClass is Drunk_2D:
        origin = Location_2D(0, 0)
    else:
        origin = Location_1D(0)
    distances = []
    if dClass is Drunk_1D:
        fieldType = Field_1D
    elif dClass is Drunk_2D:
        fieldType = Field_2D
    else:
        fieldType = Field_3D
    for t in range(numTrials):
        f = fieldType()
        f.addDrunk(molecule, origin)
        distances.append(round(walk(f, molecule,
                                    numSteps), 1))
    return distances

def drunkTest(walkLengths, numTrials, dClass):
    """Assumes walkLengths a sequence of ints >= 0
         numTrials an int > 0, dClass a subclass of Drunk
       For each number of steps in walkLengths, runs
       simWalks with numTrials walks and prints results"""
    for numSteps in walkLengths:
        distances = simWalks(numSteps, numTrials, dClass)
        print(dClass.__name__, 'random walk of',
              numSteps, 'steps')
        print(' Mean =',
              round(sum(distances)/len(distances), 4))
        print(' Max =',
              max(distances), 'Min =', min(distances))

class styleIterator(object):
    def __init__(self, styles):
        self.index = 0
        self.styles = styles

    def nextStyle(self):
        result = self.styles[self.index]
        if self.index == len(self.styles) - 1:
            self.index = 0
        else:
            self.index += 1
        return result

def simDrunk(numTrials, dClass, walkLengths):
    meanDistances = []
    for numSteps in walkLengths:
        print('Starting simulation of',
              numSteps, 'steps')
        trials = simWalks(numSteps, numTrials, dClass)
        mean = sum(trials)/len(trials)
        meanDistances.append(mean)
    return meanDistances

def simAll(drunkKinds, walkLengths, numTrials):
    styleChoice = styleIterator(('m-', 'b--', 'g-.', 'k'))
    for dClass in drunkKinds:
        curStyle = styleChoice.nextStyle()
        print('Starting simulation of', dClass.__name__)
        means = simDrunk(numTrials, dClass, walkLengths)
        plt.plot(walkLengths, means, curStyle,
                   label = dClass.__name__)
    yVals = [x**0.5 for x in walkLengths]
    plt.plot(walkLengths, yVals, styleChoice.nextStyle(),
             label = 'sqrt(x)')
    plt.title('Mean Distance from Origin ('
                + str(numTrials) + ' trials)')
    plt.xlabel('Number of Steps')
    plt.ylabel('Distance from Origin')
    plt.legend(loc = 'best')
    

random.seed(1)
numSteps = np.arange(0, 2001, 50)
simAll((Drunk_1D, Drunk_2D, Drunk_3D), numSteps, 1000)
